from latticegen.latticegeneration import *
from latticegen.singularities import hexlattice_gen_singularity, singularity_shift 

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
